import asyncio
import functools
import hashlib
import os
import re
import shutil
import subprocess
from pathlib import Path

import aiohttp
from tqdm.asyncio import tqdm

from .logger import Logger


def ffmpeg_required(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        if not shutil.which("ffmpeg"):
            Logger.error("ffmpeg is not installed")
            return
        return await func(*args, **kwargs)

    return wrapper


def _hash_id(input: str) -> str:
    hash_object = hashlib.sha256(input.encode("utf-8"))
    return hash_object.hexdigest()


async def _ts_dl(url: str, path: Path, **kwargs):
    overrides = kwargs.get("overrides", False)
    headers = kwargs.get("headers", None)

    if not overrides and path.exists():
        return

    path.unlink(missing_ok=True)
    path.parent.mkdir(parents=True, exist_ok=True)

    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            if response.status != 200:
                raise Exception("Error downloading from .ts url")
            with open(path.as_posix(), "wb") as file:
                file.write(await response.read())


async def _worker_ts_dl(urls: list, dir: Path, **kwargs):
    BATCH_SIZE = 10
    IDX = 1

    with tqdm(total=len(urls)) as bar:
        for i in range(0, len(urls), BATCH_SIZE):
            urls_batch = urls[i : i + BATCH_SIZE]
            tasks = []
            for ts_url in urls_batch:
                ts_path = dir / f"{IDX}.ts"
                tasks.append(_ts_dl(ts_url, ts_path, **kwargs))
                IDX += 1

            try:
                await asyncio.gather(*tasks)
            except Exception:
                raise Exception("Error downloading ts m3u8")

            bar.update(len(urls_batch))


async def _m3u8_dl(
    url: str,
    path: str,
    tmp_dir: str = ".tmp",
    **kwargs,
) -> None:
    overrides = kwargs.get("overrides", False)
    headers = kwargs.get("headers", None)

    if not overrides and Path(path).exists():
        return

    hash = _hash_id(url)

    Path(path).unlink(missing_ok=True)
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    Path(tmp_dir).mkdir(parents=True, exist_ok=True)

    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            if response.status != 200:
                raise Exception("Error downloading m3u8")

            pattern = r"https?://[^\s]+"
            ts_urls = re.findall(pattern, await response.text())

            if not ts_urls:
                raise Exception("No ts urls found")

            dir = Path(tmp_dir) / _hash_id(url)

            await _worker_ts_dl(ts_urls, dir, **kwargs)

    ts_files = os.listdir(dir)
    ts_files = [ts for ts in ts_files if ts.endswith(".ts")]
    ts_files = sorted(ts_files, key=lambda x: int(x.split(".")[0]))
    ts_paths = [Path(hash) / ts for ts in ts_files]

    list_file = Path(tmp_dir) / f"{hash}.txt"
    with open(list_file.as_posix(), "w") as file:
        for ts_path in ts_paths:
            file.write(f"file '{ts_path.as_posix()}'\n")

    Path(path).parent.mkdir(parents=True, exist_ok=True)

    command = [
        "ffmpeg",
        "-f",
        "concat",
        "-safe",
        "0",
        "-i",
        list_file.as_posix(),
        "-c",
        "copy",
        "-y" if overrides else "-n",
        path,
    ]

    try:
        subprocess.run(
            command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        list_file.unlink(missing_ok=True)
        shutil.rmtree(dir)

    except Exception:
        raise Exception("Error converting m3u8 to mp4")


async def m3u8_dl(
    url: str,
    path: str,
    tmp_dir: str = ".tmp",
    **kwargs,
):
    # TODO: implement quality selection
    headers = kwargs.get("headers", None)

    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            if response.status != 200:
                raise Exception("Error downloading m3u8")

            pattern = r"https?://[^\s]+"
            m3u8_urls = re.findall(pattern, await response.text())

            if not m3u8_urls:
                raise Exception("No m3u8 urls found")

            await _m3u8_dl(m3u8_urls[0], path, **kwargs)



import functools
import json
from pathlib import Path

from playwright.async_api import BrowserContext, Page, async_playwright

from .collectors import get_course_title, get_draft_chapters, get_unit
from .constants import HEADERS, LOGIN_DETAILS_URL, LOGIN_URL, SESSION_FILE
from .helpers import hash_id, read_json, write_json
from .logger import Logger
from .m3u8 import m3u8_dl
from .models import TypeUnit, Unit, User
from .utils import Cache, download, progressive_scroll, slugify


def login_required(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        self = args[0]
        if not isinstance(self, AsyncPlatzi):
            Logger.error(f"{login_required.__name__} can only decorate Platzi class.")
            return
        if not self.loggedin:
            Logger.error("Login first!")
            return
        return await func(*args, **kwargs)

    return wrapper


def try_except_request(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        self = args[0]
        if not isinstance(self, AsyncPlatzi):
            Logger.error(
                f"{try_except_request.__name__} can only decorate Platzi class."
            )
            return

        try:
            return await func(*args, **kwargs)
        except Exception as e:
            if str(e):
                Logger.error(e)
        return

    return wrapper


class AsyncPlatzi:
    def __init__(self, headless=False):
        self.loggedin = False
        self.headless = headless
        self.user = None

    async def __aenter__(self):
        self._playwright = await async_playwright().start()
        self._browser = await self._playwright.chromium.launch(headless=self.headless)
        self._context = await self._browser.new_context(
            java_script_enabled=True,
            is_mobile=True,
        )

        try:
            await self._load_state()
        except Exception:
            pass

        await self._set_profile()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._context.close()
        await self._browser.close()
        await self._playwright.stop()

    @property
    async def page(self) -> Page:
        return await self._context.new_page()

    @property
    def context(self) -> BrowserContext:
        return self._context

    @try_except_request
    async def _set_profile(self) -> None:
        try:
            data = await self.get_json(LOGIN_DETAILS_URL)
            self.user = User(**data)
        except Exception:
            return

        if self.user.is_authenticated:
            self.loggedin = True
            Logger.info(f"Hi, {self.user.username}!")

    @try_except_request
    async def login(self) -> None:
        Logger.info("Please login, in the opened browser")
        Logger.info("You have to login manually, you have 2 minutes to do it")

        page = await self.page
        await page.goto(LOGIN_URL)
        try:
            avatar = await page.wait_for_selector(
                ".styles-module_Menu__Avatar__FTuh-",
                timeout=2 * 60 * 1000,
            )
            if avatar:
                self.loggedin = True
                await self._save_state()
                Logger.info("Logged in successfully")
        except Exception:
            raise Exception("Login failed")
        finally:
            await page.close()

    @try_except_request
    async def logout(self):
        SESSION_FILE.unlink(missing_ok=True)
        Logger.info("Logged out successfully")

    @try_except_request
    @login_required
    async def download(self, url: str, **kwargs):
        page = await self.page
        await page.goto(url)

        # course title
        course_title = await get_course_title(page)
        Logger.print(course_title, "[COURSE]")

        # download directory
        DL_DIR = Path("Platzi") / slugify(course_title)
        DL_DIR.mkdir(parents=True, exist_ok=True)

        # save page as mhtml
        await self.save_page(
            page,
            path=DL_DIR / "presentation.mhtml",
        )

        # iterate over chapters
        draft_chapters = await get_draft_chapters(page)
        for idx, draft_chapter in enumerate(draft_chapters, 1):
            Logger.info(f"Downloading {draft_chapter.name}")

            CHAP_DIR = DL_DIR / f"{idx:02}_{draft_chapter.slug}"
            CHAP_DIR.mkdir(parents=True, exist_ok=True)

            # iterate over units
            for jdx, draft_unit in enumerate(draft_chapter.units, 1):
                cache_hash = hash_id(draft_unit.url)
                cache_data = Cache.get(cache_hash)

                if cache_data:
                    unit = Unit.model_validate(cache_data)
                else:
                    unit = await get_unit(self.context, draft_unit.url)
                    Cache.set(cache_hash, unit.model_dump())

                file_name = f"{jdx:02}_{unit.slug}"

                # download video
                if unit.video:
                    dst = CHAP_DIR / f"{file_name}.mp4"
                    Logger.print(f"[{dst.name}]", "[DOWNLOADING]")
                    await m3u8_dl(unit.video.url, dst.as_posix(), headers=HEADERS)

                    if unit.video.subtitles_url:
                        dst = CHAP_DIR / f"{file_name}.vtt"
                        Logger.print(f"[{dst.name}]", "[DOWNLOADING]")
                        await download(unit.video.subtitles_url, dst)

                # download lecture
                if unit.type == TypeUnit.LECTURE:
                    dst = CHAP_DIR / f"{file_name}.mhtml"
                    Logger.print(f"[{dst.name}]", "[DOWNLOADING]")
                    await self.save_page(unit.url, path=dst)

                # download quiz
                if unit.type == TypeUnit.QUIZ:
                    dst = CHAP_DIR / f"{file_name}.mhtml"
                    Logger.print(f"[{dst.name}]", "[DOWNLOADING]")
                    await self.save_page(unit.url, path=dst)

            print("=" * 100)

    @try_except_request
    async def save_page(self, src: str | Page, path: str = "source.mhtml"):
        if isinstance(src, str):
            page = await self.page
            await page.goto(src)
        else:
            page = src

        await progressive_scroll(page)

        try:
            client = await page.context.new_cdp_session(page)
            response = await client.send("Page.captureSnapshot")
            with open(path, "w", encoding="utf-8", newline="\n") as file:
                file.write(response["data"])
        except Exception:
            raise Exception("Error saving page as mhtml")

        if isinstance(src, str):
            await page.close()

    @try_except_request
    async def get_json(self, url: str) -> dict:
        page = await self.page
        await page.goto(url)
        content = await page.locator("pre").first.text_content()
        await page.close()
        return json.loads(content or "{}")

    async def _save_state(self):
        cookies = await self.context.cookies()
        write_json(SESSION_FILE, cookies)

    async def _load_state(self):
        SESSION_FILE.touch()
        cookies = read_json(SESSION_FILE)
        await self.context.add_cookies(cookies)



import asyncio
import re
from pathlib import Path

import aiohttp
from playwright.async_api import Page
from unidecode import unidecode

from .constants import SESSION_DIR
from .helpers import read_json, write_json


async def progressive_scroll(
    page: Page, time: float = 3, delay: float = 0.1, steps: int = 250
):
    delta, total_time = 0.0, 0.0
    while total_time < time:
        await asyncio.sleep(delay)
        await page.mouse.wheel(0, steps)
        delta += steps
        total_time += delay


def get_course_slug(url: str) -> str:
    """
    Extracts the course slug from a Platzi course URL.

    Args:
        url (str): The Platzi course URL.

    Returns:
        str: The course slug.

    Raises:
        Exception: If the URL is not a valid Platzi course URL.
    """
    pattern = r"https://platzi\.com/cursos/([^/]+)/?"
    match = re.search(pattern, url)
    if not match:
        raise Exception("Invalid course url")
    return match.group(1)


def clean_string(text: str) -> str:
    """
    Cleans the input string by removing special characters and
    leading/trailing white spaces.

    Args:
        text (str): The input string to be cleaned.

    Returns:
        str: The cleaned string, with special characters removed and
        leading/trailing spaces stripped.
    """
    pattern = r"[ºª]|[^\w\s]"
    return re.sub(pattern, "", text).strip()


def slugify(text: str) -> str:
    """
    Slugify a string by removing special characters and
    leading/trailing white spaces, and replacing spaces with hyphens.

    Args:
        text (str): The input string to be slugified.
    Returns:
        str: The slugified string.
    """
    return unidecode(clean_string(text)).lower().replace(" ", "-")


def get_m3u8_url(content: str) -> str:
    pattern = r"https?://[^\s\"'}]+\.m3u8"
    matches = re.findall(pattern, content)

    if not matches:
        raise Exception("No m3u8 urls found")

    return matches[0]


def get_subtitles_url(content: str) -> str | None:
    pattern = r"https?://[^\s\"'}]+\.vtt"
    matches = re.findall(pattern, content)

    if not matches:
        return None

    return matches[0]


async def download(url: str, path: Path, **kwargs):
    overrides = kwargs.get("overrides", False)
    headers = kwargs.get("headers", None)

    if not overrides and path.exists():
        return

    path.unlink(missing_ok=True)
    path.parent.mkdir(parents=True, exist_ok=True)

    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            if response.status != 200:
                raise Exception(f"Error downloading file: [{path.name}]")

            with open(path.as_posix(), "wb") as file:
                async for chunk in response.content.iter_chunked(1024):
                    file.write(chunk)


class Cache:
    @classmethod
    def get(cls, id: str) -> dict | None:
        path = SESSION_DIR / f"{id}.json"
        try:
            return read_json(path.as_posix())
        except Exception:
            return None

    @classmethod
    def set(cls, id: str, content: dict) -> None:
        path = SESSION_DIR / f"{id}.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        try:
            write_json(path.as_posix(), content)
        except Exception:
            pass



from playwright.async_api import BrowserContext, Page

from .constants import PLATZI_URL
from .models import Chapter, TypeUnit, Unit, Video
from .utils import get_m3u8_url, get_subtitles_url, slugify


async def get_course_title(page: Page) -> str:
    SELECTOR = ".Hero-content-title"
    EXCEPTION = Exception("No course title found")
    try:
        title = await page.locator(SELECTOR).first.text_content()
        if not title:
            raise EXCEPTION
    except Exception:
        await page.close()
        raise EXCEPTION

    return title


async def get_draft_chapters(page: Page) -> list[Chapter]:
    SELECTOR = ".Content-feed div.ContentBlock"
    EXCEPTION = Exception("No sections found")
    try:
        locator = page.locator(SELECTOR)

        chapters: list[Chapter] = []
        for i in range(await locator.count()):
            chapter_name = await locator.nth(i).locator("h3").first.text_content()

            if not chapter_name:
                raise EXCEPTION

            block_list_locator = locator.nth(i).locator(".ContentBlock-list a")

            units: list[Unit] = []
            for j in range(await block_list_locator.count()):
                ITEM_LOCATOR = block_list_locator.nth(j)

                unit_url = await ITEM_LOCATOR.get_attribute("href")
                unit_title = await ITEM_LOCATOR.locator("h5").first.text_content()

                if not unit_url or not unit_title:
                    raise EXCEPTION

                units.append(
                    Unit(
                        type=TypeUnit.VIDEO,
                        title=unit_title,
                        url=PLATZI_URL + unit_url,
                        slug=slugify(unit_title),
                    )
                )

            chapters.append(
                Chapter(
                    name=chapter_name,
                    slug=slugify(chapter_name),
                    units=units,
                )
            )

    except Exception as e:
        await page.close()
        raise EXCEPTION from e

    return chapters


async def get_unit(context: BrowserContext, url: str) -> Unit:
    TYPE_SELECTOR = ".VideoPlayer"
    TITLE_SELECTOR = ".MaterialDesktopHeading_MaterialDesktopHeading-info__title__DaYr2"
    EXCEPTION = Exception("Could not collect unit data")

    if "/quiz/" in url:
        return Unit(
            url=url,
            title="Quiz",
            type=TypeUnit.QUIZ,
            slug=slugify("Quiz"),
        )

    try:
        page = await context.new_page()
        await page.goto(url)

        title = await page.locator(TITLE_SELECTOR).first.text_content()

        if not title:
            raise EXCEPTION

        if await page.locator(TYPE_SELECTOR).count() == 0:
            type = TypeUnit.LECTURE
            video = None

        else:
            content = await page.content()
            type = TypeUnit.VIDEO
            video = Video(
                url=get_m3u8_url(content),
                subtitles_url=get_subtitles_url(content),
            )

        return Unit(
            url=url,
            title=title,
            type=type,
            video=video,
            slug=slugify(title),
        )

    except Exception:
        raise EXCEPTION

    finally:
        await page.close()



from pathlib import Path

import platformdirs

APP_NAME = "Platzi"
SESSION_DIR = Path(platformdirs.user_data_dir(APP_NAME))
SESSION_FILE = SESSION_DIR / "state.json"

LOGIN_URL = "https://platzi.com/login"
LOGIN_DETAILS_URL = "https://api.platzi.com/api/v1/components/headerv2/user/"


PLATZI_URL = "https://platzi.com"
HOME_URL = PLATZI_URL + "/home"

USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36"

HEADERS = {
    "User-Agent": USER_AGENT,
    "Referer": HOME_URL,
}


# --- Session directory ---
SESSION_FILE.parent.mkdir(parents=True, exist_ok=True)



import hashlib
import json


def read_json(path: str) -> dict:
    with open(path, "r", encoding="utf-8") as file:
        return json.load(file)


def write_json(path: str, data: dict) -> None:
    with open(path, "w", encoding="utf-8") as file:
        json.dump(data, file, indent=4, ensure_ascii=False)


def write_file(path: str, content: str) -> None:
    with open(path, "w", encoding="utf-8") as file:
        file.write(content)


def hash_id(input_string: str) -> str:
    hash_object = hashlib.sha256(input_string.encode("utf-8"))
    return hash_object.hexdigest()
